@@ -1,4 +1,4 @@
-# Copyright (C) 2011-2016 2ndQuadrant Italia Srl
+# Copyright (C) 2011-2017 2ndQuadrant Limited
 #
 # This file is part of Barman.
 #
@@ -26,6 +26,7 @@ through Ssh).
 A BackupExecutor is invoked by the BackupManager for backup operations.
 """
 
+import datetime
 import logging
 import os
 import re
@@ -46,7 +47,8 @@ from barman.exceptions import (CommandFailedException, DataTransferFailure,
 from barman.fs import UnixRemoteCommand
 from barman.infofile import BackupInfo
 from barman.remote_status import RemoteStatusMixin
-from barman.utils import mkpath, with_metaclass
+from barman.utils import (human_readable_timedelta, mkpath, total_seconds,
+                          with_metaclass)
 
 _logger = logging.getLogger(__name__)
 
@@ -69,6 +71,8 @@ class BackupExecutor(with_metaclass(ABCMeta, RemoteStatusMixin)):
         self.config = backup_manager.config
         self.strategy = None
         self._mode = mode
+        self.copy_start_time = None
+        self.copy_end_time = None
 
         # Holds the action being executed. Used for error messages.
         self.current_action = None
@@ -158,6 +162,24 @@ class BackupExecutor(with_metaclass(ABCMeta, RemoteStatusMixin)):
                                 "has been removed",
                                 wal_name, self.config.name)
 
+    def _start_backup_copy_message(self, backup_info):
+        """
+        Output message for backup start
+
+        :param barman.infofile.BackupInfo backup_info: backup information
+        """
+        output.info("Copying files for %s", backup_info.backup_id)
+
+    def _stop_backup_copy_message(self, backup_info):
+        """
+        Output message for backup end
+
+        :param barman.infofile.BackupInfo backup_info: backup information
+        """
+        output.info("Copy done (time: %s)",
+                    human_readable_timedelta(datetime.timedelta(
+                        seconds=backup_info.copy_stats['copy_time'])))
+
 
 def _parse_ssh_command(ssh_command):
     """
@@ -298,9 +320,9 @@ class PostgresBackupExecutor(BackupExecutor):
 
             # Start the copy
             self.current_action = "copying files"
-            output.info("Copying files.")
+            self._start_backup_copy_message(backup_info)
             self.backup_copy(backup_info)
-            output.info("Copy done.")
+            self._stop_backup_copy_message(backup_info)
             self.strategy.stop_backup(backup_info)
 
             # If this is the first backup, purge eventually unused WAL files
@@ -316,17 +338,18 @@ class PostgresBackupExecutor(BackupExecutor):
         :param CheckStrategy check_strategy: the strategy for the management
              of the results of the various checks
         """
+        check_strategy.init_check('pg_basebackup')
         remote_status = self.get_remote_status()
 
         # Check for the presence of pg_basebackup
         check_strategy.result(
-            self.config.name, 'pg_basebackup',
-            remote_status['pg_basebackup_installed'])
+            self.config.name, remote_status['pg_basebackup_installed'])
 
         # remote_status['pg_basebackup_compatible'] is None if
         # pg_basebackup cannot be executed and False if it is
         # not compatible.
         hint = None
+        check_strategy.init_check('pg_basebackup compatible')
         if not remote_status['pg_basebackup_compatible']:
             pg_version = 'Unknown'
             basebackup_version = 'Unknown'
@@ -338,7 +361,7 @@ class PostgresBackupExecutor(BackupExecutor):
                 pg_version, basebackup_version
             )
         check_strategy.result(
-            self.config.name, 'pg_basebackup compatible',
+            self.config.name,
             remote_status['pg_basebackup_compatible'], hint=hint)
 
         # Skip further checks if the postgres connection doesn't work.
@@ -348,6 +371,7 @@ class PostgresBackupExecutor(BackupExecutor):
         if postgres is None or postgres.server_txt_version is None:
             return
 
+        check_strategy.init_check('pg_basebackup supports tablespaces mapping')
         # We can't backup a cluster with tablespaces if the tablespace
         # mapping option is not available in the installed version
         # of pg_basebackup.
@@ -387,7 +411,6 @@ class PostgresBackupExecutor(BackupExecutor):
 
         check_strategy.result(
             self.config.name,
-            'pg_basebackup supports tablespaces mapping',
             check_result,
             hint=hint
         )
@@ -482,6 +505,9 @@ class PostgresBackupExecutor(BackupExecutor):
         backup_dest = backup_info.get_data_directory()
         dest_dirs = [backup_dest]
 
+        # Store the start time
+        self.copy_start_time = datetime.datetime.now()
+
         # Manage tablespaces, we need to handle them now in order to
         # be able to relocate them inside the
         # destination directory of the basebackup
@@ -504,6 +530,10 @@ class PostgresBackupExecutor(BackupExecutor):
         if remote_status['pg_basebackup_bwlimit']:
             bandwidth_limit = self.config.bandwidth_limit
 
+        # Make sure we are not wasting precious PostgreSQL resources
+        # for the whole duration of the copy
+        self.server.close()
+
         pg_basebackup = PgBaseBackup(
             connection=self.server.streaming,
             destination=backup_dest,
@@ -527,6 +557,32 @@ class PostgresBackupExecutor(BackupExecutor):
             raise DataTransferFailure.from_command_error(
                 'pg_basebackup', e, msg)
 
+        # Store the end time
+        self.copy_end_time = datetime.datetime.now()
+
+        # Store statistics about the copy
+        copy_time = total_seconds(self.copy_end_time - self.copy_start_time)
+        backup_info.copy_stats = {
+            'copy_time': copy_time,
+            'total_time': copy_time,
+        }
+
+        # Check for the presence of configuration files outside the PGDATA
+        external_config = backup_info.get_external_config_files()
+        if any(external_config):
+            msg = ("pg_basebackup does not copy the PostgreSQL "
+                   "configuration files that reside outside PGDATA. "
+                   "Please manually backup the following files:\n"
+                   "\t%s\n" %
+                   "\n\t".join(ecf.path for ecf in external_config))
+            # Show the warning only if the EXTERNAL_CONFIGURATION option
+            # is not specified in the backup_options.
+            if (BackupOptions.EXTERNAL_CONFIGURATION
+                    not in self.config.backup_options):
+                output.warning(msg)
+            else:
+                _logger.debug(msg)
+
     def _retry_handler(self, dest_dirs, command, args, kwargs,
                        attempt, exc):
         """
@@ -570,6 +626,10 @@ class PostgresBackupExecutor(BackupExecutor):
             # chmod 0700 octal
             os.chmod(dest_dir, 448)
 
+    def _start_backup_copy_message(self, backup_info):
+        output.info("Starting backup copy via pg_basebackup for %s",
+                    backup_info.backup_id)
+
 
 class SshBackupExecutor(with_metaclass(ABCMeta, BackupExecutor)):
     """
@@ -676,9 +736,9 @@ class SshBackupExecutor(with_metaclass(ABCMeta, BackupExecutor)):
 
             # Start the copy
             self.current_action = "copying files"
-            output.info("Copying files.")
+            self._start_backup_copy_message(backup_info)
             self.backup_copy(backup_info)
-            output.info("Copy done.")
+            self._stop_backup_copy_message(backup_info)
 
             # Try again to purge eventually unused WAL files. At this point
             # the begin_wal value is surely known. Doing it twice is safe
@@ -687,11 +747,12 @@ class SshBackupExecutor(with_metaclass(ABCMeta, BackupExecutor)):
         except:
             # we do not need to do anything here besides re-raising the
             # exception. It will be handled in the external try block.
+            output.error("The backup has failed %s", self.current_action)
             raise
         else:
             self.current_action = "issuing stop of the backup"
-            output.info("Asking PostgreSQL server to finalize the backup.")
         finally:
+            output.info("Asking PostgreSQL server to finalize the backup.")
             try:
                 self.strategy.stop_backup(backup_info)
             except BaseException:
@@ -707,7 +768,7 @@ class SshBackupExecutor(with_metaclass(ABCMeta, BackupExecutor)):
         :param CheckStrategy check_strategy: the strategy for the management
              of the results of the various checks
         """
-
+        check_strategy.init_check('ssh')
         hint = "PostgreSQL server"
         cmd = None
         minimal_ssh_output = None
@@ -720,16 +781,16 @@ class SshBackupExecutor(with_metaclass(ABCMeta, BackupExecutor)):
                 hint = str(e).strip()
 
         # Output the result
-        check_strategy.result(self.config.name, 'ssh', cmd is not None, hint)
+        check_strategy.result(self.config.name, cmd is not None, hint=hint)
 
         # Check if the communication channel is "clean"
         if minimal_ssh_output:
+            check_strategy.init_check('ssh output clean')
             check_strategy.result(
                 self.config.name,
-                'ssh output clean',
                 False,
-                "the configured ssh_command must not add anything "
-                "to the remote command output")
+                hint="the configured ssh_command must not add anything to "
+                     "the remote command output")
 
         # If SSH works but PostgreSQL is not responding
         if (cmd is not None and
@@ -741,6 +802,7 @@ class SshBackupExecutor(with_metaclass(ABCMeta, BackupExecutor)):
             )
             # Look for the latest backup in the catalogue
             if last_backup:
+                check_strategy.init_check('backup_label')
                 # Get PGDATA and build path to 'backup_label'
                 backup_label = os.path.join(last_backup.pgdata,
                                             'backup_label')
@@ -750,9 +812,7 @@ class SshBackupExecutor(with_metaclass(ABCMeta, BackupExecutor)):
                 if exists:
                     hint = "Check that the PostgreSQL server is up " \
                            "and no 'backup_label' file is in PGDATA."
-                    check_strategy.result(self.config.name,
-                                          'backup_label', False,
-                                          hint)
+                    check_strategy.result(self.config.name, False, hint=hint)
 
         try:
             # Invoke specific checks for the backup strategy
@@ -800,7 +860,7 @@ class SshBackupExecutor(with_metaclass(ABCMeta, BackupExecutor)):
                     archive_dir = os.path.join(
                         self.server.postgres.get_setting('data_directory'),
                         'pg_xlog', 'archive_status')
-                    out = str(cmd.list_dir_content('-t', archive_dir))
+                    out = str(cmd.list_dir_content(archive_dir, ['-t']))
                     for line in out.splitlines():
                         if line.endswith('.done'):
                             name = line[:-5]
@@ -811,6 +871,14 @@ class SshBackupExecutor(with_metaclass(ABCMeta, BackupExecutor)):
             _logger.warn("Error retrieving PostgreSQL status: %s", e)
         return remote_status
 
+    def _start_backup_copy_message(self, backup_info):
+        number_of_workers = self.config.parallel_jobs
+        message = "Starting backup copy via rsync/SSH for %s" % (
+            backup_info.backup_id,)
+        if number_of_workers > 1:
+            message += " (%s jobs)" % number_of_workers
+        output.info(message)
+
 
 class RsyncBackupExecutor(SshBackupExecutor):
     """
@@ -878,6 +946,10 @@ class RsyncBackupExecutor(SshBackupExecutor):
             backup_info.backup_id)
         safe_horizon = None
         reuse_backup = None
+
+        # Store the start time
+        self.copy_start_time = datetime.datetime.now()
+
         if previous_backup:
             # safe_horizon is a tz-aware timestamp because BackupInfo class
             # ensures that property
@@ -896,6 +968,7 @@ class RsyncBackupExecutor(SshBackupExecutor):
             safe_horizon=safe_horizon,
             retry_times=self.config.basebackup_retry_times,
             retry_sleep=self.config.basebackup_retry_sleep,
+            workers=self.config.parallel_jobs,
         )
 
         # List of paths to be excluded by the PGDATA copy
@@ -1002,18 +1075,29 @@ class RsyncBackupExecutor(SshBackupExecutor):
             raise DataTransferFailure.from_command_error(
                 'rsync', e, msg)
 
+        # Store the end time
+        self.copy_end_time = datetime.datetime.now()
+
+        # Store statistics about the copy
+        backup_info.copy_stats = controller.statistics()
+
         # Check for any include directives in PostgreSQL configuration
         # Currently, include directives are not supported for files that
         # reside outside PGDATA. These files must be manually backed up.
         # Barman will emit a warning and list those files
         if any(included_config_files):
-            output.warning(
-                "The usage of include directives is not supported "
-                "for files that reside outside PGDATA.\n"
-                "Please manually backup the following files:\n"
-                "\t%s\n",
-                "\n\t".join(icf.path for icf in included_config_files)
-            )
+            msg = ("The usage of include directives is not supported "
+                   "for files that reside outside PGDATA.\n"
+                   "Please manually backup the following files:\n"
+                   "\t%s\n" %
+                   "\n\t".join(icf.path for icf in included_config_files))
+            # Show the warning only if the EXTERNAL_CONFIGURATION option
+            # is not specified in the backup_options.
+            if (BackupOptions.EXTERNAL_CONFIGURATION
+                    not in self.config.backup_options):
+                output.warning(msg)
+            else:
+                _logger.debug(msg)
 
     def _reuse_path(self, previous_backup_info, tablespace=None):
         """
@@ -1120,6 +1204,10 @@ class BackupStrategy(with_metaclass(ABCMeta, object)):
         # Set server version
         backup_info.set_attribute('version', server.postgres.server_version)
 
+        # Set XLOG segment size
+        backup_info.set_attribute('xlog_segment_size',
+                                  server.postgres.xlog_segment_size)
+
         # Set configuration files location
         cf = server.postgres.get_configuration_files()
         for key in cf:
@@ -1156,7 +1244,8 @@ class BackupStrategy(with_metaclass(ABCMeta, object)):
             start_info = start_info.copy()
             start_info.update(xlog.location_to_xlogfile_name_offset(
                 start_info['location'],
-                start_info['timeline']))
+                start_info['timeline'],
+                backup_info.xlog_segment_size))
 
         # If file_name and file_offset are available, use them
         if (start_info.get('file_name') is not None and
@@ -1195,7 +1284,8 @@ class BackupStrategy(with_metaclass(ABCMeta, object)):
                 timeline = backup_info.timeline
             stop_info.update(xlog.location_to_xlogfile_name_offset(
                 stop_info['location'],
-                timeline))
+                timeline,
+                backup_info.xlog_segment_size))
 
         backup_info.set_attribute('end_time', stop_info['timestamp'])
         backup_info.set_attribute('end_xlog', stop_info['location'])
@@ -1231,7 +1321,7 @@ class BackupStrategy(with_metaclass(ABCMeta, object)):
         backup_info.set_attribute('begin_xlog', wal_info.group(1))
         backup_info.set_attribute('begin_wal', wal_info.group(2))
         backup_info.set_attribute('begin_offset', xlog.parse_lsn(
-            wal_info.group(1)) % xlog.XLOG_SEG_SIZE)
+            wal_info.group(1)) % backup_info.xlog_segment_size)
         backup_info.set_attribute('begin_time', dateutil.parser.parse(
             start_time.group(1)))
 
@@ -1287,28 +1377,20 @@ class PostgresBackupStrategy(BackupStrategy):
 
         # Get the current xlog position
         postgres = self.executor.server.postgres
-        current_xlog_info = postgres.current_xlog_info.copy()
-        self._backup_info_from_stop_location(backup_info, current_xlog_info)
+        current_xlog_info = postgres.current_xlog_info
+        if current_xlog_info:
+            self._backup_info_from_stop_location(
+                backup_info, current_xlog_info)
 
         # Ask PostgreSQL to switch to another XLOG file. This is needed
         # to archive the transaction log file containing the backup
         # end position, which is required to recover from the backup.
-        postgres = self.executor.server.postgres
         try:
             postgres.switch_xlog()
         except PostgresIsInRecovery:
             # Skip switching XLOG if a standby server
             pass
 
-        # Check for the presence of configuration files outside the PGDATA
-        external_config = backup_info.get_external_config_files()
-        if any(external_config):
-            output.warning("pg_basebackup does not copy the PostgreSQL "
-                           "configuration files that reside outside PGDATA. "
-                           "Please manually backup the following files:\n"
-                           "\t%s\n",
-                           "\n\t".join(ecf.path for ecf in external_config))
-
 
 class ExclusiveBackupStrategy(BackupStrategy):
     """
@@ -1378,15 +1460,16 @@ class ExclusiveBackupStrategy(BackupStrategy):
              of the results of the various checks
         """
         # Make sure PostgreSQL is not in recovery (i.e. is a master)
+        check_strategy.init_check('not in recovery')
         if self.executor.server.postgres:
             is_in_recovery = self.executor.server.postgres.is_in_recovery
             if not is_in_recovery:
                 check_strategy.result(
-                    self.executor.config.name, 'not in recovery', True)
+                    self.executor.config.name, True)
             else:
                 check_strategy.result(
-                    self.executor.config.name, 'not in recovery', False,
-                    'cannot perform exclusive backup on a standby')
+                    self.executor.config.name, False,
+                    hint='cannot perform exclusive backup on a standby')
 
 
 class ConcurrentBackupStrategy(BackupStrategy):
@@ -1486,11 +1569,6 @@ class ConcurrentBackupStrategy(BackupStrategy):
         self.current_action = "writing backup label"
         self._write_backup_label(backup_info)
 
-        # Write the tablespaces map only if at least a tablespace is present
-        if backup_info.tablespaces:
-            self.current_action = "writing tablespace map"
-            self._write_tablespace_map(backup_info)
-
         # Ask PostgreSQL to switch to another XLOG file. This is needed
         # to archive the transaction log file containing the backup
         # end position, which is required to recover from the backup.
@@ -1536,6 +1614,7 @@ class ConcurrentBackupStrategy(BackupStrategy):
         :param CheckStrategy check_strategy: the strategy for the management
              of the results of the various checks
         """
+        check_strategy.init_check('pgespresso extension')
         postgres = self.executor.server.postgres
         try:
             # We execute this check only if the postgres connection is non None
@@ -1543,14 +1622,12 @@ class ConcurrentBackupStrategy(BackupStrategy):
             # there is a native API for concurrent backups.
             if postgres and postgres.server_version < 90600:
                 if postgres.has_pgespresso:
-                    check_strategy.result(self.executor.config.name,
-                                          'pgespresso extension', True)
+                    check_strategy.result(self.executor.config.name, True)
                 else:
-                    check_strategy.result(self.executor.config.name,
-                                          'pgespresso extension', False,
-                                          'required for concurrent backups on '
-                                          'PostgreSQL %s' %
-                                          postgres.server_major_version)
+                    check_strategy.result(self.executor.config.name, False,
+                                          hint='required for concurrent '
+                                               'backups on PostgreSQL %s' %
+                                               postgres.server_major_version)
         except PostgresConnectionError:
             # Skip the check if the postgres connection doesn't work.
             # We assume that this error condition will be reported by
@@ -1567,6 +1644,7 @@ class ConcurrentBackupStrategy(BackupStrategy):
         """
         postgres = self.executor.server.postgres
         start_info = postgres.start_concurrent_backup(label)
+        postgres.allow_reconnect = False
         self._backup_info_from_start_location(backup_info, start_info)
 
     def _concurrent_stop_backup(self, backup_info):
@@ -1578,5 +1656,6 @@ class ConcurrentBackupStrategy(BackupStrategy):
         """
         postgres = self.executor.server.postgres
         stop_info = postgres.stop_concurrent_backup()
+        postgres.allow_reconnect = True
         backup_info.set_attribute('backup_label', stop_info['backup_label'])
         self._backup_info_from_stop_location(backup_info, stop_info)
