@@ -1,4 +1,4 @@
-# Copyright (C) 2011-2016 2ndQuadrant Italia Srl
+# Copyright (C) 2011-2017 2ndQuadrant Limited
 #
 # This file is part of Barman.
 #
@@ -507,6 +507,7 @@ class RecoveryExecutor(object):
             safe_horizon=safe_horizon,
             retry_times=self.config.basebackup_retry_times,
             retry_sleep=self.config.basebackup_retry_sleep,
+            workers=self.config.parallel_jobs,
         )
 
         # Dictionary for paths to be excluded from rsync
@@ -551,10 +552,12 @@ class RecoveryExecutor(object):
             dst=dest_prefix + dest,
             bwlimit=self.config.get_bwlimit(),
             exclude=[
-                '/pg_xlog/*',
                 '/pg_log/*',
+                '/pg_xlog/*',
+                '/postmaster.pid',
                 '/recovery.conf',
-                '/postmaster.pid'],
+                '/tablespace_map',
+            ],
             exclude_and_protect=exclude_and_protect,
             item_class=controller.PGDATA_CLASS
         )
