@@ -1,4 +1,4 @@
-# Copyright (C) 2011-2016 2ndQuadrant Italia Srl
+# Copyright (C) 2011-2017 2ndQuadrant Limited
 #
 # This file is part of Barman.
 #
@@ -20,7 +20,6 @@ This module represents a Server.
 Barman is able to manage multiple servers.
 """
 
-import itertools
 import logging
 import os
 import shutil
@@ -28,6 +27,7 @@ import sys
 import time
 from collections import namedtuple
 from contextlib import contextmanager
+from glob import glob
 from tempfile import NamedTemporaryFile
 
 import barman
@@ -55,7 +55,8 @@ from barman.postgres import PostgreSQLConnection, StreamingConnection
 from barman.process import ProcessManager
 from barman.remote_status import RemoteStatusMixin
 from barman.retention_policies import RetentionPolicyFactory
-from barman.utils import human_readable_timedelta, pretty_size, timeout
+from barman.utils import (human_readable_timedelta, is_power_of_two,
+                          pretty_size, timeout)
 from barman.wal_archiver import (FileWalArchiver, StreamingWalArchiver,
                                  WalArchiver)
 
@@ -78,7 +79,12 @@ class CheckStrategy(object):
     NON_CRITICAL_CHECKS = ['minimum redundancy requirements',
                            'backup maximum age',
                            'failed backups',
-                           'archiver errors']
+                           'archiver errors',
+                           'empty incoming directory',
+                           'empty streaming directory',
+                           'incoming WALs directory',
+                           'streaming WALs directory',
+                           ]
 
     def __init__(self, ignore_checks=NON_CRITICAL_CHECKS):
         """
@@ -89,17 +95,34 @@ class CheckStrategy(object):
         self.ignore_list = ignore_checks
         self.check_result = []
         self.has_error = False
+        self.running_check = None
 
-    def result(self, server_name, check, status, hint=None):
+    def init_check(self, check_name):
+        """
+        Mark in the debug log when barman starts the execution of a check
+
+        :param str check_name: the name of the check that is starting
+        """
+        self.running_check = check_name
+        _logger.debug("Starting check: %s" % check_name)
+
+    def _check_name(self, check):
+        if not check:
+            check = self.running_check
+        assert check
+        return check
+
+    def result(self, server_name, status, hint=None, check=None):
         """
         Store the result of a check (with no output).
         Log any check result (error or debug level).
 
         :param str server_name: the server is being checked
-        :param str check: the check name
         :param bool status: True if succeeded
         :param str,None hint: hint to print if not None:
+        :param str,None check: the check name
         """
+        check = self._check_name(check)
         if not status:
             # If the name of the check is not in the filter list,
             # treat it as a blocking error, then notify the error
@@ -122,6 +145,7 @@ class CheckStrategy(object):
         # Store the result and does not output anything
         result = self.CheckResult(server_name, check, status)
         self.check_result.append(result)
+        self.running_check = None
 
 
 class CheckOutputStrategy(CheckStrategy):
@@ -138,18 +162,20 @@ class CheckOutputStrategy(CheckStrategy):
         """
         super(CheckOutputStrategy, self).__init__(ignore_checks=())
 
-    def result(self, server_name, check, status, hint=None):
+    def result(self, server_name, status, hint=None, check=None):
         """
-        Output Strategy constructor
+        Store the result of a check.
+        Log any check result (error or debug level).
+        Output the result to the user
 
         :param str server_name: the server being checked
         :param str check: the check name
         :param bool status: True if succeeded
         :param str,None hint: hint to print if not None:
         """
-        # Call the basic method
+        check = self._check_name(check)
         super(CheckOutputStrategy, self).result(
-            server_name, check, status, hint)
+            server_name, status, hint, check)
         # Send result to output
         output.result('check', server_name, check, status, hint)
 
@@ -429,8 +455,11 @@ class Server(RemoteStatusMixin):
         except TimeoutError:
             # The check timed out.
             # Add a failed entry to the check strategy for this.
-            check_strategy.result(self.config.name, 'check timeout', False,
-                                  'barman check command timed out')
+            _logger.debug("Check command timed out executing '%s' check"
+                          % check_strategy.running_check)
+            check_strategy.result(self.config.name, False,
+                                  hint='barman check command timed out',
+                                  check='check timeout')
 
     def check_archive(self, check_strategy):
         """
@@ -439,7 +468,7 @@ class Server(RemoteStatusMixin):
         :param CheckStrategy check_strategy: the strategy for the management
              of the results of the various checks
         """
-
+        check_strategy.init_check("WAL archive")
         # Make sure that WAL archiving has been setup
         # XLOG_DB needs to exist and its size must be > 0
         # NOTE: we do not need to acquire a lock in this phase
@@ -452,8 +481,66 @@ class Server(RemoteStatusMixin):
         # NOTE: This check needs to be only visible if it fails
         if xlogdb_empty:
             check_strategy.result(
-                self.config.name, 'WAL archive', False,
-                'please make sure WAL shipping is setup')
+                self.config.name, False,
+                hint='please make sure WAL shipping is setup')
+
+        # Check the number of wals in the incoming directory
+        self._check_wal_queue(check_strategy,
+                              'incoming',
+                              'archiver')
+
+        # Check the number of wals in the streaming directory
+        self._check_wal_queue(check_strategy,
+                              'streaming',
+                              'streaming_archiver')
+
+    def _check_wal_queue(self, check_strategy, dir_name, archiver_name):
+        """
+        Check if one of the wal queue directories beyond the
+        max file threshold
+        """
+        # Read the wal queue location from the configuration
+        config_name = "%s_wals_directory" % dir_name
+        assert hasattr(self.config, config_name)
+        incoming_dir = getattr(self.config, config_name)
+
+        # Check if the archiver is enabled
+        assert hasattr(self.config, archiver_name)
+        enabled = getattr(self.config, archiver_name)
+
+        # Inspect the wal queue directory
+        file_count = len(glob(os.path.join(incoming_dir, '*')))
+        max_incoming_wal = self.config.max_incoming_wals_queue
+
+        # Subtract one from the count because of .partial file inside the
+        # streaming directory
+        if dir_name == 'streaming':
+            file_count -= 1
+
+        # If this archiver is disabled, check the number of files in the
+        # corresponding directory.
+        # If the directory is NOT empty, fail the check and warn the user.
+        # NOTE: This check is visible only when it fails
+        check_strategy.init_check("empty %s directory" % dir_name)
+        if not enabled:
+            if file_count > 0:
+                check_strategy.result(
+                    self.config.name, False,
+                    hint="'%s' must be empty when %s=off"
+                         % (incoming_dir, archiver_name))
+            # No more checks are required if the archiver
+            # is not enabled
+            return
+
+        # At this point if max_wals_count is none,
+        # means that no limit is set so we just need to return
+        if max_incoming_wal is None:
+            return
+        check_strategy.init_check("%s WALs directory" % dir_name)
+        if file_count > max_incoming_wal:
+            msg = 'there are too many WALs in queue: ' \
+                  '%s, max %s' % (file_count, max_incoming_wal)
+            check_strategy.result(self.config.name, False, hint=msg)
 
     def check_postgres(self, check_strategy):
         """
@@ -462,24 +549,30 @@ class Server(RemoteStatusMixin):
         :param CheckStrategy check_strategy: the strategy for the management
              of the results of the various checks
         """
+        check_strategy.init_check('PostgreSQL')
         # Take the status of the remote server
         remote_status = self.get_remote_status()
         if remote_status.get('server_txt_version'):
-            check_strategy.result(self.config.name, 'PostgreSQL', True)
+            check_strategy.result(self.config.name, True)
         else:
-            check_strategy.result(self.config.name, 'PostgreSQL', False)
+            check_strategy.result(self.config.name, False)
             return
         # Check for superuser privileges in PostgreSQL
         if remote_status.get('is_superuser') is not None:
+            check_strategy.init_check('is_superuser')
             if remote_status.get('is_superuser'):
                 check_strategy.result(
-                    self.config.name, 'superuser', True)
+                    self.config.name, True)
             else:
                 check_strategy.result(
-                    self.config.name, 'not superuser', False,
-                    'superuser privileges for PostgreSQL connection required')
+                    self.config.name, False,
+                    hint='superuser privileges for PostgreSQL '
+                         'connection required',
+                    check='not superuser'
+                )
 
         if 'streaming_supported' in remote_status:
+            check_strategy.init_check("PostgreSQL streaming")
             hint = None
 
             # If a streaming connection is available,
@@ -489,22 +582,24 @@ class Server(RemoteStatusMixin):
             elif not remote_status['streaming_supported']:
                 hint = ('Streaming connection not supported'
                         ' for PostgreSQL < 9.2')
-            check_strategy.result(self.config.name, 'PostgreSQL streaming',
-                                  remote_status.get('streaming'), hint)
+            check_strategy.result(self.config.name,
+                                  remote_status.get('streaming'), hint=hint)
         # Check wal_level parameter: must be different from 'minimal'
         # the parameter has been introduced in postgres >= 9.0
         if 'wal_level' in remote_status:
+            check_strategy.init_check("wal_level")
             if remote_status['wal_level'] != 'minimal':
                 check_strategy.result(
-                    self.config.name, 'wal_level', True)
+                    self.config.name, True)
             else:
                 check_strategy.result(
-                    self.config.name, 'wal_level', False,
-                    "please set it to a higher level than 'minimal'")
+                    self.config.name, False,
+                    hint="please set it to a higher level than 'minimal'")
 
         # Check the presence and the status of the configured replication slot
         # This check will be skipped if `slot_name` is undefined
         if self.config.slot_name:
+            check_strategy.init_check("replication slot")
             slot = remote_status['replication_slot']
             # The streaming_archiver is enabled
             if self.config.streaming_archiver is True:
@@ -512,39 +607,37 @@ class Server(RemoteStatusMixin):
                 if not remote_status['replication_slot_support']:
                     check_strategy.result(
                         self.config.name,
-                        'replication slot',
                         False,
-                        "slot_name parameter set but "
-                        "PostgreSQL server is too old (%s < 9.4)" % (
-                            remote_status['server_txt_version']))
+                        hint="slot_name parameter set but PostgreSQL server "
+                             "is too old (%s < 9.4)" %
+                             remote_status['server_txt_version'])
                 # Replication slots are supported
                 else:
                     # The slot is not present
                     if slot is None:
                         check_strategy.result(
-                            self.config.name, 'replication slot', False,
-                            "replication slot '%s' doesn't exist. "
-                            "Please execute "
-                            "'barman receive-wal --create-slot %s'" %
-                            (self.config.slot_name, self.config.name))
+                            self.config.name, False,
+                            hint="replication slot '%s' doesn't exist. "
+                                 "Please execute 'barman receive-wal "
+                                 "--create-slot %s'" % (self.config.slot_name,
+                                                        self.config.name))
                     else:
                         # The slot is present but not initialised
                         if slot.restart_lsn is None:
                             check_strategy.result(
-                                self.config.name, 'replication slot', False,
-                                "slot '%s' not initialised: "
-                                "is 'receive-wal' running?" %
-                                self.config.slot_name)
+                                self.config.name, False,
+                                hint="slot '%s' not initialised: is "
+                                     "'receive-wal' running?" %
+                                     self.config.slot_name)
                         # The slot is present but not active
                         elif slot.active is False:
                             check_strategy.result(
-                                self.config.name, 'replication slot', False,
-                                "slot '%s' not active: "
-                                "is 'receive-wal' running?" %
-                                self.config.slot_name)
+                                self.config.name, False,
+                                hint="slot '%s' not active: is "
+                                     "'receive-wal' running?" %
+                                     self.config.slot_name)
                         else:
                             check_strategy.result(self.config.name,
-                                                  'replication slot',
                                                   True)
             else:
                 # If the streaming_archiver is disabled and the slot_name
@@ -564,11 +657,10 @@ class Server(RemoteStatusMixin):
                         # Warn the user
                         check_strategy.result(
                             self.config.name,
-                            'replication slot',
                             True,
-                            "WARNING: slot '%s' is %s but not required "
-                            "by the current config" % (
-                                self.config.slot_name, slot_status))
+                            hint="WARNING: slot '%s' is %s but not required "
+                                 "by the current config" % (
+                                  self.config.slot_name, slot_status))
 
     def _make_directories(self):
         """
@@ -588,14 +680,15 @@ class Server(RemoteStatusMixin):
         :param CheckStrategy check_strategy: the strategy for the management
              of the results of the various checks
         """
+        check_strategy.init_check("directories")
         if not self.config.disabled:
             try:
                 self._make_directories()
             except OSError as e:
-                check_strategy.result(self.config.name, 'directories', False,
+                check_strategy.result(self.config.name, False,
                                       "%s: %s" % (e.filename, e.strerror))
             else:
-                check_strategy.result(self.config.name, 'directories', True)
+                check_strategy.result(self.config.name, True)
 
     def check_configuration(self, check_strategy):
         """
@@ -605,8 +698,9 @@ class Server(RemoteStatusMixin):
         :param CheckStrategy check_strategy: the strategy for the management
              of the results of the various checks
         """
+        check_strategy.init_check('configuration')
         if len(self.config.msg_list):
-            check_strategy.result(self.config.name, 'configuration', False)
+            check_strategy.result(self.config.name, False)
             for conflict_paths in self.config.msg_list:
                 output.info("\t\t%s" % conflict_paths)
 
@@ -617,14 +711,12 @@ class Server(RemoteStatusMixin):
         :param CheckStrategy check_strategy: the strategy for the management
              of the results of the various checks
         """
+        check_strategy.init_check("retention policy settings")
         if (self.config.retention_policy and
                 not self.enforce_retention_policies):
-            check_strategy.result(self.config.name,
-                                  'retention policy settings', False,
-                                  'see log')
+            check_strategy.result(self.config.name, False, hint='see log')
         else:
-            check_strategy.result(self.config.name,
-                                  'retention policy settings', True)
+            check_strategy.result(self.config.name, True)
 
     def check_backup_validity(self, check_strategy):
         """
@@ -633,6 +725,7 @@ class Server(RemoteStatusMixin):
         :param CheckStrategy check_strategy: the strategy for the management
              of the results of the various checks
         """
+        check_strategy.init_check('backup maximum age')
         # first check: check backup maximum age
         if self.config.last_backup_maximum_age is not None:
             # get maximum age information
@@ -641,16 +734,16 @@ class Server(RemoteStatusMixin):
 
             # format the output
             check_strategy.result(
-                self.config.name, 'backup maximum age',
-                backup_age[0],
-                "interval provided: %s, latest backup age: %s" % (
+                self.config.name, backup_age[0],
+                hint="interval provided: %s, latest backup age: %s" % (
                     human_readable_timedelta(
                         self.config.last_backup_maximum_age), backup_age[1]))
         else:
             # last_backup_maximum_age provided by the user
             check_strategy.result(
-                self.config.name, 'backup maximum age', True,
-                "no last_backup_maximum_age provided")
+                self.config.name,
+                True,
+                hint="no last_backup_maximum_age provided")
 
     def check_archiver_errors(self, check_strategy):
         """
@@ -659,7 +752,7 @@ class Server(RemoteStatusMixin):
         :param CheckStrategy check_strategy: the strategy for the management
              of the results of the check
         """
-
+        check_strategy.init_check('archiver errors')
         if os.path.isdir(self.config.errors_directory):
             errors = os.listdir(self.config.errors_directory)
         else:
@@ -667,9 +760,8 @@ class Server(RemoteStatusMixin):
 
         check_strategy.result(
             self.config.name,
-            "archiver errors",
             len(errors) == 0,
-            WalArchiver.summarise_error_files(errors)
+            hint=WalArchiver.summarise_error_files(errors)
         )
 
     def status_postgres(self):
@@ -1121,7 +1213,7 @@ class Server(RemoteStatusMixin):
 
             # evaluation of compression ratio for basebackup WAL files
             wal_info['wal_theoretical_size'] = \
-                wal_info['wal_num'] * float(xlog.XLOG_SEG_SIZE)
+                wal_info['wal_num'] * float(backup_info.xlog_segment_size)
             try:
                 wal_info['wal_compression_ratio'] = 1 - (
                     wal_info['wal_size'] /
@@ -1131,7 +1223,8 @@ class Server(RemoteStatusMixin):
 
             # evaluation of compression ratio of WAL files
             wal_info['wal_until_next_theoretical_size'] = \
-                wal_info['wal_until_next_num'] * float(xlog.XLOG_SEG_SIZE)
+                wal_info['wal_until_next_num'] * \
+                float(backup_info.xlog_segment_size)
             try:
                 wal_info['wal_until_next_compression_ratio'] = 1 - (
                     wal_info['wal_until_next_size'] /
@@ -1176,9 +1269,19 @@ class Server(RemoteStatusMixin):
         :param int|None peek: if defined list the next N WAL file
         """
 
+        # If used through SSH identify the client to add it to logs
+        source_suffix = ''
+        ssh_connection = os.environ.get('SSH_CONNECTION')
+        if ssh_connection:
+            # The client IP is the first value contained in `SSH_CONNECTION`
+            # which contains four space-separated values: client IP address,
+            # client port number, server IP address, and server port number.
+            source_suffix = ' (SSH host: %s)' % (ssh_connection.split()[0],)
+
         # Sanity check
         if not xlog.is_any_xlog_file(wal_name):
-            output.error("'%s' is not a valid wal file name", wal_name)
+            output.error("'%s' is not a valid wal file name%s",
+                         wal_name, source_suffix)
             return
 
         # If peek is requested we only output a list of files
@@ -1188,17 +1291,53 @@ class Server(RemoteStatusMixin):
             # we cannot guess the names of the following WAL files.
             # So ``wal_name`` is the only possible result, if exists.
             if xlog.is_wal_file(wal_name):
-                wal_peek_list = itertools.islice(
-                    xlog.generate_segment_names(wal_name), peek)
+                # We can't know what was the segment size of PostgreSQL WAL
+                # files at backup time. Because of this, we generate all
+                # the possible names for a WAL segment, and then we check
+                # if the requested one is included.
+                wal_peek_list = xlog.generate_segment_names(wal_name)
             else:
                 wal_peek_list = [wal_name]
-            # Output the content of wal_peek_list until we find a missing file
-            for wal_peek_name in wal_peek_list:
+
+            # Output the content of wal_peek_list until we have displayed
+            # enough files or find a missing file
+            count = 0
+            while count < peek:
+                try:
+                    wal_peek_name = next(wal_peek_list)
+                except StopIteration:
+                    # No more item in wal_peek_list
+                    break
+
                 wal_peek_file = self.get_wal_full_path(wal_peek_name)
-                # If ``wal_peek_file`` doesn't exist, stop the process
-                if not os.path.exists(wal_peek_file):
-                    return
-                output.info(wal_peek_name, log=False)
+
+                # If the next WAL file is found, output the name
+                # and continue to the next one
+                if os.path.exists(wal_peek_file):
+                    count += 1
+                    output.info(wal_peek_name, log=False)
+                    continue
+
+                # If ``wal_peek_file`` doesn't exist, check if we need to
+                # look in the following segment
+                tli, log, seg = xlog.decode_segment_name(wal_peek_name)
+
+                # If `seg` is not a power of two, it is not possible that we
+                # are at the end of a WAL group, so we are done
+                if not is_power_of_two(seg):
+                    break
+
+                # This is a possible WAL group boundary, let's try the
+                # following group
+                seg = 0
+                log += 1
+
+                # Install a new generator from the start of the next segment.
+                # If the file doesn't exists we will terminate because
+                # zero is not a power of two
+                wal_peek_name = xlog.encode_segment_name(tli, log, seg)
+                wal_peek_list = xlog.generate_segment_names(wal_peek_name)
+
             # Do not output anything else
             return
 
@@ -1207,8 +1346,8 @@ class Server(RemoteStatusMixin):
 
         # Check for file existence
         if not os.path.exists(wal_file):
-            output.error("WAL file '%s' not found in server '%s'",
-                         wal_name, self.config.name)
+            output.error("WAL file '%s' not found in server '%s'%s",
+                         wal_name, self.config.name, source_suffix)
             return
 
         # If an output directory was provided write the file inside it
@@ -1217,14 +1356,19 @@ class Server(RemoteStatusMixin):
             destination_path = os.path.join(output_directory, wal_name)
             try:
                 destination = open(destination_path, 'w')
-                output.info("Writing WAL '%s' for server '%s' into '%s' file",
-                            wal_name, self.config.name, destination_path)
+                output.info(
+                    "Writing WAL '%s' for server '%s' into '%s' file%s",
+                    wal_name, self.config.name, destination_path,
+                    source_suffix)
             except IOError as e:
-                output.error("Unable to open '%s' file: %s" %
-                             destination_path, e)
+                output.error("Unable to open '%s' file%s: %s",
+                             destination_path, source_suffix, e)
                 return
         else:
             destination = sys.stdout
+            _logger.info(
+                "Writing WAL '%s' for server '%s' to standard output%s",
+                wal_name, self.config.name, source_suffix)
 
         # Get a decompressor for the file (None if not compressed)
         wal_compressor = self.backup_manager.compression_manager \
@@ -1691,7 +1835,7 @@ class Server(RemoteStatusMixin):
             output.error(
                 "invalid xlog segment name %r\n"
                 "HINT: Please run \"barman rebuild-xlogdb %s\" "
-                "to solve this issue" %
+                "to solve this issue",
                 str(e), self.config.name)
             output.close_and_exit()
 
