@@ -1,4 +1,4 @@
-# Copyright (C) 2011-2016 2ndQuadrant Italia Srl
+# Copyright (C) 2011-2017 2ndQuadrant Limited
 #
 # This file is part of Barman.
 #
@@ -203,6 +203,22 @@ def human_readable_timedelta(timedelta):
     return human
 
 
+def total_seconds(timedelta):
+    """
+    Compatibility method because the total_seconds method has been introduced
+    in Python 2.7
+
+    :param timedelta: a timedelta object
+    :rtype: float
+    """
+    if hasattr(timedelta, 'total_seconds'):
+        return timedelta.total_seconds()
+    else:
+        return (timedelta.microseconds +
+                (timedelta.seconds + timedelta.days * 24 * 3600) * 10**6
+                ) / 10.0**6
+
+
 def which(executable, path=None):
     """
     This method is useful to find if a executable is present into the
@@ -348,3 +364,18 @@ def timeout(timeout_duration):
         # Reset the signal
         signal.alarm(0)
         signal.signal(signal.SIGALRM, signal.SIG_DFL)
+
+
+def is_power_of_two(number):
+    """
+    Check if a number is a power of two or not
+    """
+
+    # This is a fast method to check for a power of two.
+    #
+    # A power of two has this structure:  100000 (one or more zeroes)
+    # This is the same number minus one:  011111 (composed by ones)
+    # This is the bitwise and:            000000
+    #
+    # This is true only for every power of two
+    return number != 0 and (number & (number-1)) == 0
